"""
Send a message by email to one or more recipients (by SMTP or sendmail).
"""

# Copyright (c) 2009-2024, UChicago Argonne, LLC.  See LICENSE file.

import os
import sys

SMTP_TIMEOUT = 10


class MailerError(Exception):
    pass


def sendMail_sendmail(
    subject, message, recipients, sendmail_cfg, sender=None, logger=None
):
    """
    Send an email message using sendmail (linux only).

    :param str subject: short text for email subject
    :param str message: full text of email body
    :param [str] recipients: list of email addresses to receive the message
    :param dict sendmail_cfg: such as returned from :mod:`PvMail.ini_config.Config.get`

        :user: required - (*str*) username to for sendmail (or similar) program

    :param str sender: "From" address, if *None* use *smtp_cfg['user']* value
    :param obj logger: optional message logging method

    EXAMPLE::

        >>> import PvMail.ini_config
        >>> sendmail_cfg = PvMail.ini_config.Config().get()
        >>> recipients = ['joe@gmail.com', 'sally@example.org']
        >>> subject = 'sendmail test message'
        >>> message = PvMail.ini_config.__doc__
        >>> sendMail_sendmail(subject, message, recipients, sendmail_cfg)

    """

    if sys.platform not in ("linux2"):
        raise MailerError(f"Cannot use this method on {sys.platform=!r}")

    sender = sender or sendmail_cfg["user"]
    if isinstance(recipients, str):
        recipients = [
            recipients,
        ]

    def _sendmail_handler(email_program, from_addr, recipients, subject, message):
        to_addr = str(" ".join(recipients))
        mail_command = "%s -F %s -t %s" % (email_program, from_addr, to_addr)
        mail_message = [
            mail_command,
        ]
        for who in recipients:
            mail_message.append("To: " + who)
        mail_message.append("Subject: " + subject)
        mail_message.append(message)
        cmd = """cat << +++ | %s\n+++""" % "\n".join(mail_message)
        return mail_command, cmd

    def _mail_handler(email_program, from_addr, recipients, subject, message):
        # to_addr = str(" ".join(recipients))
        # mail_command = "%s %s" % (email_program, to_addr)
        # mail_message = "%s\n%s" % (subject, message)
        # # TODO: needs to do THIS
        # """
        # cat /tmp/message.txt | mail jemian@anl.gov
        # """
        # # cmd = '''echo %s | %s''' % (mail_message, mail_command)
        # # return mail_command, cmd    # lgtm [py/unreachable-statement]
        raise MailerError("code needs improvement here")

    cmd = None
    for email_program, handler in (
        ("/usr/lib/sendmail", _sendmail_handler),
        ("/usr/bin/sendmail", _sendmail_handler),
        ("/usr/bin/mail", _mail_handler),
    ):
        if os.path.exists(email_program):
            results = handler(email_program, sender, recipients, subject, message)
            mail_command, cmd = results
            break

    if cmd is None:
        raise MailerError("Cannot find mail transport agent for sendmail.")

    if logger is not None:
        logger("sending email to: " + str(recipients))
        logger("email program: " + email_program)
        logger("mail command: " + mail_command)
        logger("email From: " + sender)

    try:
        if logger is not None:
            logger("email command:\n" + cmd)
        if os.path.exists(email_program):
            os.popen(cmd)  # send the message
        else:
            if logger is not None:
                logger("email program (%s) does not exist" % email_program)
    except Exception as reason:
        # err_msg = traceback.format_exc()
        final_msg = f"{cmd=!r}\ntraceback: {reason}"
        logger(final_msg)
        raise MailerError(final_msg)

    if logger is not None:
        logger("sendmail sent")


def sendMail_SMTP(subject, message, recipients, smtp_cfg, sender=None, logger=None):
    """
    send email message through SMTP server

    :param str subject: short text for email subject
    :param str message: full text of email body
    :param [str] recipients: list of email addresses to receive the message
    :param dict smtp_cfg: such as returned from :mod:`PvMail.ini_config.Config.get`

        :server:               required - (*str*) SMTP server
        :user:                 required - (*str*) username to login to SMTP server
        :port:                 optional - (*str*) SMTP port
        :password:             optional - (*str*) password for username
        :connection_security:  optional - (*str*) **STARTTLS** (the only choice, if specified)

    :param str sender: "From" address, if *None* use *smtp_cfg['user']* value

    EXAMPLE::

        >>> import PvMail.ini_config
        >>> smtp_cfg = PvMail.ini_config.Config().get()
        >>> recipients = ['joe@gmail.com', 'sally@example.org']
        >>> subject = 'SMTP test message'
        >>> message = PvMail.ini_config.__doc__
        >>> sendMail_SMTP(subject, message, recipients, smtp_cfg)

    """
    import email
    import smtplib

    host = smtp_cfg.get("server", None)
    if host is None:
        raise MailerError("must define an SMTP host to be used")
    port = smtp_cfg.get("port", None)
    username = smtp_cfg.get("user", None)
    if username is None:
        raise MailerError("must define a username for the SMTP server")
    if sender is None:
        sender = username
    password = smtp_cfg.get("password", None)
    connection_security = smtp_cfg.get("connection_security", None)
    if connection_security not in (None, "STARTTLS"):
        msg = "connection_security must be: STARTTLS or not defined, found: "
        raise MailerError(msg + connection_security)

    msg = email.message.Message()
    if isinstance(recipients, str):
        msg["To"] = recipients
    else:
        for who in recipients:
            msg["To"] = who
    msg["From"] = sender
    msg["Subject"] = subject
    msg.set_payload(message)

    if logger is not None:
        logger("sending email to: " + str(recipients))
        logger("SMTP server: " + host)
        logger("SMTP user: " + username)
        logger("email From: " + sender)

    smtpserver = smtplib.SMTP(timeout=SMTP_TIMEOUT)
    # smtpserver.set_debuglevel(1)
    if port is None:
        smtpserver.connect(host)
    else:
        smtpserver.connect(host, int(port))
    if logger is not None:
        logger("SMTP connected")

    smtpserver.ehlo()
    if smtp_cfg.get("connection_security", None) == "STARTTLS":
        smtpserver.starttls()
        smtpserver.ehlo()
        if logger is not None:
            logger("SMTP STARTTLS")

    if password is not None:
        smtpserver.login(username, password)
        if logger is not None:
            logger("SMTP authenticated")

    smtpserver.sendmail(username, recipients, str(msg))
    smtpserver.close()
    if logger is not None:
        logger("SMTP complete")


def send_message(subject, message, recipients, config):
    """
    send an email message

    :param str subject: short text for email subject
    :param str message: full text of email body
    :param [str] recipients: list of email addresses to receive the message
    :param dict config: such as returned from :mod:`PvMail.ini_config.Config`
    """
    email_agent_dict = dict(sendmail=sendMail_sendmail, SMTP=sendMail_SMTP)
    agent = email_agent_dict[config.mail_transfer_agent]
    agent(subject, message, recipients, config.get())


def main():
    """
    User on-demand test of the mailer module and configuration.
    """
    import argparse

    from . import DOCS_URL
    from . import __version__
    from . import ini_config

    doc = f"Test the email sender from PvMail {__version__}"
    parser = argparse.ArgumentParser(description=doc)
    msg = "email address(es), whitespace-separated if more than one"
    parser.add_argument("recipient", action="store", nargs="+", help=msg, default="")
    results = parser.parse_args()
    print(doc)

    cfg = ini_config.Config()
    print("Using configuration from: " + cfg.ini_file)
    print("Using user name: " + cfg.get()["user"])

    recipients = results.recipient
    print("Sending email(s) to: " + str(" ".join(recipients)))
    print("mail transfer agent: " + cfg.mail_transfer_agent)

    subject = "PvMail mailer test message: " + cfg.mail_transfer_agent
    message = f"{doc}\nFor more help, see: {DOCS_URL}"

    send_message(subject, message, recipients, cfg)



"""
utility routines for PvMail

Copyright (c) 2009-2024, UChicago Argonne, LLC.  See LICENSE file.
"""

import pathlib

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


def get_pkg_file_path(fname):
    """return the absolute path to the named file in the package directory"""
    vpath = pathlib.Path(__file__).parent
    vfile = vpath / fname
    if not vfile.exists():
        raise FileNotFoundError(f"file does not exist: {vfile}")
    return vfile


def read_resource_file(fname):
    """return contents of the named file in the package directory"""
    return open(get_pkg_file_path(fname)).read()



"""
unit tests for the spec module
"""

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2017, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE, distributed with this software.
# -----------------------------------------------------------------------------


import datetime
import socket
import sys
import unittest  # TODO: refactor to pytest

from PvMail import ini_config
from PvMail import mailer


class Test(unittest.TestCase):

    def setUp(self):
        self.cfg = ini_config.Config()

    #     def tearDown(self):
    #         pass

    def test_sendmail(self):
        # this should fail on windows
        recipients = ["jemian@anl.gov"]
        subject = "pvMail test using sendmail"
        # message = "\n".join(["%s\t%s" % (k, v) for k, v in sorted(os.environ.items())])
        message = "pvMail mailer test_sendmail"
        mailer.sendMail_sendmail(subject, message, recipients)

    def test_sendmail_2(self):
        # this should fail on windows
        recipients = ["jemian", "prjemian"]
        message = ""
        message += "host: %s\n" % socket.gethostname()
        message += "date: %s (UNIX, not PV)\n" % datetime.datetime.now()
        message += "program: %s\n" % sys.argv[0]
        message += "recipients: %s\n" % ", ".join(recipients)
        subject = "pvMail mailer test_sendmail_2"
        mailer.sendMail_sendmail(subject, message, recipients)

    def test_smtp(self):
        subject = "pvMail mailer test_smtp"
        message_text = subject
        agent = self.cfg.get()
        mailer.sendMail_SMTP(
            [
                "prjemian@gmail.com",
            ],
            message_text,
            subject=subject,
            recipient_name=agent["user"],
            sender_email=agent["user"],
            sender_password=agent["password"],
            simulation=False,
            smtp_server=agent["server"],
        )


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    # unittest.main()

    suite = unittest.TestLoader().loadTestsFromTestCase(Test)
    unittest.TextTestRunner(verbosity=2).run(suite)



"""
unit tests for the spec module
"""

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2017, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE, distributed with this software.
# -----------------------------------------------------------------------------


import logging
import time
import unittest  # TODO: refactor to pytest

from PvMail import cli


class Test(unittest.TestCase):

    def setUp(self):
        pass

    def test_sendmail(self):
        logging.basicConfig(filename=cli.LOG_FILE, level=logging.INFO)
        cli.logger("startup")
        pvm = cli.PvMail()
        pvm.recipients = ["prjemian@gmail.com"]
        pvm.triggerPV = "pvMail:trigger"
        pvm.messagePV = "pvMail:message"
        retry_interval_s = 0.05
        end_time = time.time() + 60
        report_time = time.time() + 5.0
        pvm.do_start()
        while time.time() < end_time:
            if time.time() > report_time:
                report_time = time.time() + 5.0
                cli.logger(
                    "time remaining: %.1f seconds ..." % (end_time - time.time())
                )
            time.sleep(retry_interval_s)
        pvm.do_stop()


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    # unittest.main()

    suite = unittest.TestLoader().loadTestsFromTestCase(Test)
    unittest.TextTestRunner(verbosity=2).run(suite)



"""
Email list model for Qt's MVC QListView
"""

from PyQt5 import QtCore


class EmailListModel(QtCore.QAbstractListModel):
    def __init__(self, input_list, parent=None, *args):
        """
        data model for GUI: supports list of email addresses

        :param [str] input_list: list of email addresses
        :param QWidget parent: view widget for this data model
        """
        super(EmailListModel, self).__init__(parent, *args)
        self.listdata = input_list

    def rowCount(self, parent=QtCore.QModelIndex()):
        return len(self.listdata)

    def data(self, index, role):
        if index.isValid():
            if role == QtCore.Qt.DisplayRole:
                return QtCore.QVariant(self.listdata[index.row()])
            elif role == QtCore.Qt.EditRole:
                return QtCore.QVariant(self.listdata[index.row()])
            else:
                return QtCore.QVariant()

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        if role == QtCore.Qt.EditRole:
            row = index.row()
            self.listdata[row] = str(value)
            if row + 1 == len(self.listdata):
                # always keep a blank item at end to grow the list
                self.listdata.append("")
            self.dataChanged.emit(index, index)
            return True
        return False

    def flags(self, index):
        defaultFlags = QtCore.QAbstractItemModel.flags(self, index)

        if index.isValid():
            return (
                defaultFlags
                | QtCore.Qt.ItemIsEditable
                | QtCore.Qt.ItemIsDragEnabled
                | QtCore.Qt.ItemIsDropEnabled
            )

        else:
            return defaultFlags | QtCore.Qt.ItemIsDropEnabled

    def reset(self):
        self.listdata = []



"""Configuration of PvMail package."""

PROJECT = "PvMail"
DESCRIPTION = "Watch an EPICS PV. Send email when it changes from 0 to 1."
DOCS_URL = "https://BCDA-APS.github.io/pvMail"
AUTHORS = [
    "Pete Jemian",
    "Kurt Goetze",
]
COPYRIGHT = "Copyright (c) 2009-2024, UChicago Argonne, LLC."
LICENSE = "LICENSE"

try:
    from setuptools_scm import get_version

    __version__ = get_version(root="../..", relative_to=__file__)
    del get_version
except (LookupError, ModuleNotFoundError):
    from importlib.metadata import version

    __version__ = version(PROJECT)



"""
Handle application configuration settings in a .ini file.

To identify the configuration file (and create if it does not exist already)::

    [joeuser] $ pvMail_mail_config_file
    /home/joeuser/.pvMail/pvMail.ini

"""

# Copyright (c) 2009-2024, UChicago Argonne, LLC.  See LICENSE file.

import datetime
import os
import sys
from configparser import ConfigParser

APPLICATION = "pvMail"
INI_FILE = "pvMail.ini"


class Unknown_MTA(Exception):
    pass


class NoConfigFile(Exception):
    pass


class Config(object):
    def __init__(self):
        if "PVMAIL_INI_FILE" in os.environ:
            self.ini_file = os.path.abspath(os.environ["PVMAIL_INI_FILE"])
            self.ini_dir = os.path.dirname(self.ini_file)
        else:
            op_sys_config_dir_env = dict(posix="HOME", nt="APPDATA")[os.name]
            dir_prefix = dict(posix=".", nt="")[os.name]
            base_dir = os.environ.get(op_sys_config_dir_env, None) or "."

            self.ini_dir = os.path.join(base_dir, dir_prefix + APPLICATION)
            self.ini_file = os.path.join(self.ini_dir, INI_FILE)

        self.mail_transfer_agent = "sendmail"
        if sys.platform not in ("linux2"):
            # only use sendmail on linux systems
            self.mail_transfer_agent = "SMTP"

        u1 = os.environ.get("LOGNAME", None)
        u2 = os.environ.get("USERNAME", None)
        u3 = "joeuser"
        username = u1 or u2 or u3

        self.agent_db = dict(
            sendmail=dict(user=username),
            SMTP=dict(
                user=username,
                password="keep_this_private",
                server="smtp.server.org",
                port="465",
                connection_security="STARTTLS",
            ),
        )

        try:
            self.read()
        except NoConfigFile:
            self.write()
            self.read()

    def setAgent(self, agent):
        """choose the mail transfer agent"""
        if agent in self.agent_db:
            self.mail_transfer_agent = agent
        else:
            raise Unknown_MTA(str(agent))

    def read(self):
        """read the configuration file"""
        if not os.path.exists(self.ini_file):
            raise NoConfigFile(str(self.ini_file))

        config = ConfigParser()
        config.read(self.ini_file)

        self.mail_transfer_agent = config.get("mailer", "mail_transfer_agent")

        for section in config.sections():
            if section not in ("header", "mailer"):
                for option in config.options(section):
                    if section not in self.agent_db:
                        self.agent_db[section] = {}
                    self.agent_db[section][option] = config.get(section, option)

    def write(self):
        """
        (re)write the configuration file
        """
        config = ConfigParser()
        config.add_section("header")
        config.set("header", "application", "pvMail")
        config.set("header", "written", str(datetime.datetime.now()))

        config.add_section("mailer")
        config.set("mailer", "mail_transfer_agent", self.mail_transfer_agent)

        for section in self.agent_db:
            config.add_section(section)
            for k, v in self.agent_db[section].items():
                config.set(section, k, v)

        if not os.path.exists(self.ini_dir):
            os.mkdir(self.ini_dir)
        config.write(open(self.ini_file, "w"))

    def get(self):
        """
        return the chosen configuration dictionary
        """
        return self.agent_db[self.mail_transfer_agent]


def main():
    con = Config()
    print(con.ini_file)
    print(len(con.agent_db), " configuration(s) described")



"""
===================================
pvMail: just the GUI
===================================

Run the Graphical User Interface for PvMail using
PyQt from a .ui file with the uic subpackage.
"""

# Copyright (c) 2009-2024, UChicago Argonne, LLC.  See LICENSE file.

import datetime
import os
import pathlib
import sys

from pydm.widgets.label import PyDMLabel
from pydm.widgets.line_edit import PyDMLineEdit
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets
from PyQt5 import uic

from . import AUTHORS
from . import COPYRIGHT
from . import DESCRIPTION
from . import DOCS_URL
from . import __version__ as VERSION
from . import cli
from . import ini_config
from . import utils
from .email_model import EmailListModel

pyqtSignal = QtCore.pyqtSignal

WINDOW_TITLE = "pvMail"
COLOR_ON = "lightgreen"
COLOR_OFF = "lightred"
COLOR_DEFAULT = "#eee"
LOG_REDISPLAY_DELAY_MS = 1000
PATH_RESOURCES = pathlib.Path(__file__).parent / "resources"
PATH_GUI_UI = utils.get_pkg_file_path(PATH_RESOURCES / "gui.ui")
PATH_ABOUT_UI = utils.get_pkg_file_path(PATH_RESOURCES / "about.ui")

PYDM_TOOLTIP_TEXT = "\n".join(
    [
        "$(pv_value.address)",
        "$(pv_value.TIME)",
        "connected=$(pv_value.connection)",
    ]
)


class PvMailSignalDef(QtCore.QObject):
    """Define the signals used to communicate between the threads."""

    EPICS_monitor = pyqtSignal(object)


class PvMail_GUI(object):
    """GUI used for pvMail, based on Qt (PyQt5 & PyDM)."""

    def __init__(
        self, ui_file=None, logger=None, logfile=None, config=None, *args, **kw
    ):
        """make this class callable from pvMail application"""
        self.ui = uic.loadUi(utils.get_pkg_file_path(ui_file or PATH_GUI_UI))

        self.ui.history.clear()
        self.logger = logger
        self.logfile = logfile
        self.config = config or ini_config.Config()

        self.setStatus("starting")

        self.email_address_model = EmailListModel([], self.ui)
        self.pvmail = None
        self.watching = False

        # menu item handlers
        self.ui.actionSend_test_email.triggered.connect(self.doSendTestMessage)
        self.ui.actionE_xit.triggered.connect(self.doClose)
        self.ui.actionAbout_pvMail.triggered.connect(self.doAbout)

        # button handlers
        self.ui.w_btn_run.clicked.connect(self.doRun)
        self.ui.w_btn_stop.clicked.connect(self.doStop)

        # the list of email recipients
        self.ui.emails.setModel(self.email_address_model)

        # adjust dynamic labels
        self.ui.config_file_name.setText(config.ini_file)
        self.ui.log_file_name.setText(str(logfile))
        self.ui.w_running_stopped.setText("stopped")

        self.ui.l_pv_trigger = QtWidgets.QLabel("trigger")
        self.ui.pv_trigger = PyDMLabel()
        self.ui.pv_trigger.setToolTip(PYDM_TOOLTIP_TEXT)

        self.ui.messagePV.returnPressed.connect(self.setMessageChannel)
        self.ui.triggerPV.returnPressed.connect(self.setTriggerChannel)

        self.ui.formLayout.addRow(self.ui.l_pv_trigger, self.ui.pv_trigger)
        self.triggerSignal = PvMailSignalDef()
        self.triggerSignal.EPICS_monitor.connect(self.onTrigger_gui_thread)

        self.ui.l_pv_message = QtWidgets.QLabel("message")
        self.ui.pv_message = PyDMLineEdit()
        self.ui.pv_message.setToolTip(PYDM_TOOLTIP_TEXT)
        self.ui.pv_message.setReadOnly(True)
        self.ui.formLayout.addRow(self.ui.l_pv_message, self.ui.pv_message)
        self.messageSignal = PvMailSignalDef()
        self.messageSignal.EPICS_monitor.connect(self.onMessage_gui_thread)

        self.setStatus("ready")

    def show(self):
        self.ui.show()

    def doAbout(self, *args, **kw):
        # TODO:  add a check for a new version #5
        # based on a comparison with the PyPI version:
        # https://pypi.python.org/pypi/PvMail
        #
        # Comparing with the VERSION file will show the development trunk
        # which is not always the production release version.
        # https://raw.githubusercontent.com/BCDA-APS/pvMail/master/src/PvMail/VERSION
        #
        about = uic.loadUi(utils.get_pkg_file_path(PATH_ABOUT_UI))
        about.title.setText(f"PvMail  {VERSION}")
        about.description.setText(DESCRIPTION)
        about.authors.setText(", ".join(AUTHORS))
        about.copyright.setText(COPYRIGHT)

        pb = QtWidgets.QPushButton(DOCS_URL, clicked=self.doUrl)
        about.verticalLayout_main.addWidget(pb)

        # TODO: provide control to show the license

        # feed the status message
        msg = f"About: PvMail, v{VERSION}, PID={os.getpid()}"
        self.setStatus(msg)
        about.show()
        about.exec_()

    def doUrl(self):
        self.setStatus("opening documentation URL in default browser")
        service = QtGui.QDesktopServices()
        url = QtCore.QUrl(DOCS_URL)
        service.openUrl(url)

    def doClose(self, *args, **kw):
        self.setStatus("application exit requested")
        self.ui.close()

    def doRun(self, *args, **kw):
        self.setStatus("<Run> button pressed")
        if self.watching:
            self.setStatus("already watching")
        else:
            self.pvmail = cli.PvMail(self.config)

            # acquire information, validate it first
            msg_pv = str(self.getMessagePV())
            if len(msg_pv) == 0:
                self.setStatus("must give a message PV name")
                return
            trig_pv = str(self.getTriggerPV())
            if len(trig_pv) == 0:
                self.setStatus("must give a trigger PV name")
                return
            if len(self.email_list) == 0:
                self.setStatus("need at least one email address for list of recipients")
                return

            # report connection failure and abort
            for key, pv in dict(message=msg_pv, trigger=trig_pv).items():
                tc = self.pvmail.testConnect(pv, timeout=cli.CONNECTION_TEST_TIMEOUT)
                if tc is False:
                    msg = "could not connect to %s PV: %s" % (key, pv)
                    self.setStatus(msg)
                    self.pvmail = None
                    return

            self.pvmail.triggerPV = trig_pv
            self.pvmail.messagePV = msg_pv

            for obj in (self.ui.messagePV, self.ui.triggerPV):
                obj.setReadOnly(True)
                obj.setStyleSheet(f"background-color: {COLOR_DEFAULT}")
            self.ui.w_running_stopped.setStyleSheet(f"background-color: {COLOR_ON}")
            self.ui.pv_message.setReadOnly(True)

            self.pvmail.recipients = self.email_list
            self.setStatus("trigger PV: " + self.pvmail.triggerPV)
            self.setStatus("message PV: " + self.pvmail.messagePV)
            self.setStatus("recipients: " + "  ".join(self.email_list))
            try:
                self.pvmail.do_start()
            except Exception as reason:
                self.setStatus(str(reason))
                return
            self.ui.w_running_stopped.setText("running")
            self.ui.w_running_stopped.setStyleSheet(
                (f"background-color: {COLOR_ON}" "; qproperty-alignment: AlignCenter")
            )
            self.watching = True
            self.setStatus("Watching for email triggers.")

    def doStop(self, *args, **kw):
        if not self.watching:
            self.setStatus("not watching now")
        else:
            self.setStatus("<Stop> button pressed")
            self.pvmail.do_stop()
            self.ui.w_running_stopped.setText("stopped")
            for obj in (self.ui.messagePV, self.ui.triggerPV):
                obj.setStyleSheet("background-color: white")
                obj.setReadOnly(False)
            self.ui.w_running_stopped.setStyleSheet(
                f"background-color: {COLOR_DEFAULT}"
            )
            self.ui.pv_message.setReadOnly(False)

            self.setStatus("Not watching, no emails will be sent.")
            self.pvmail = None
            self.watching = False

    def doSendTestMessage(self):
        from . import mailer

        self.setStatus("requested a test email")
        subject = "PvMail mailer test message: " + self.config.mail_transfer_agent
        message = f"This is a test of the PvMail mailer, v{VERSION}"
        msg_pv = str(self.getMessagePV())
        if self.watching:
            message += "\n\n" + self.pvmail.message
        message += "\n\n triggerPV = " + str(self.getTriggerPV())
        message += "\n messagePV = " + msg_pv
        message += f"\n\n For more help, see: {DOCS_URL}"
        mailer.send_message(subject, message, self.email_list, self.config)
        self.setStatus("sent a test email")

    def appendEmailList(self, email_addr):
        self.email_address_model.listdata.append(email_addr)
        self.setStatus("added email: " + email_addr)

    @property
    def email_list(self):
        """the list of email addresses with empty items removed"""
        recipients = [v for v in self.email_address_model.listdata if len(v) > 0]
        return recipients

    @email_list.setter
    def email_list(self, recipients):
        self.email_address_model.reset()
        # remove blanks from the list, append blank at end for user to append
        recipients = (" ".join(recipients)).strip().split() + [""]
        for v in recipients:
            self.appendEmailList(v)

    def getMessagePV(self):
        return self.ui.messagePV.text()

    def onMessage_pv_thread(self, value=None, *args, **kw):
        self.messageSignal.EPICS_monitor.emit(value)  # threadsafe update of the widget

    def onMessage_gui_thread(self, value):
        self.setStatus("message: %s" % str(value))
        if self.ui.pv_message.text() != value:
            self.ui.pv_message.text_cache = value
            self.ui.pv_message.SetText()

    def setMessagePV(self, messagePV):
        self.ui.messagePV.setText(str(messagePV))
        self.setMessageChannel()
        self.setStatus("set message PV: " + messagePV)

    def setMessageChannel(self):
        self.setWidgetChannel(self.ui.pv_message, self.getMessagePV())

    def getTriggerPV(self):
        return self.ui.triggerPV.text()

    def onTrigger_pv_thread(self, value=None, char_value=None, *args, **kw):
        self.triggerSignal.EPICS_monitor.emit(value)  # threadsafe update of the widget

    def onTrigger_gui_thread(self, value):
        self.setStatus("trigger: %s" % str(value))
        color = {0: COLOR_DEFAULT, 1: COLOR_ON}[value]
        self.ui.pv_trigger.SetBackgroundColor(color)
        if self.ui.pv_trigger.text() != str(value):
            self.ui.pv_trigger.text_cache = str(value)
            self.ui.pv_trigger.SetText()
        if value == 1:
            # delayed update, wait for sending the email to be logged
            QtCore.QTimer.singleShot(LOG_REDISPLAY_DELAY_MS, self.logfile_to_history)

    def setTriggerPV(self, triggerPV):
        self.ui.triggerPV.setText(str(triggerPV))
        self.setTriggerChannel()
        self.setStatus("set trigger PV: " + triggerPV)

    def setWidgetChannel(self, widget, pv):
        pv = pv.strip()
        if len(pv) == 0:
            pv = "-none-"
        widget.set_channel(f"ca://{pv}")
        widget.check_enable_state()

    def setTriggerChannel(self):
        self.setWidgetChannel(self.ui.pv_trigger, self.getTriggerPV())

    def setStatus(self, message):
        ts = str(datetime.datetime.now())
        self.ui.statusbar.showMessage(str(message))
        if hasattr(self.ui, "history"):
            if self.logger is None or self.logfile is None:
                self.ui.history.append(ts + "  " + message)
            else:
                self.logger(message)
                # update with log file contents
                self.logfile_to_history()
            self.ui.history.ensureCursorVisible()

    def logfile_to_history(self):
        # update history with logfile contents
        self.ui.history.clear()
        self.ui.history.append(open(self.logfile, "r").read())


def main(triggerPV, messagePV, recipients, logger=None, logfile=None, config=None):
    app = QtWidgets.QApplication(sys.argv)
    config = config or ini_config.Config()
    gui = PvMail_GUI(logger=logger, logfile=logfile, config=config)

    gui.setStatus("PID: " + str(os.getpid()))
    if logfile is not None:
        logfile = pathlib.Path(logfile)
    if logfile is not None and logfile.exists():
        gui.ui.history.append(open(logfile, "r").read())
        gui.setStatus(f"log file: {logfile}")
    gui.setStatus("email configuration file: " + config.ini_file)
    gui.setStatus("email agent: " + config.mail_transfer_agent)
    gui.setMessagePV(messagePV)
    gui.setTriggerPV(triggerPV)
    gui.email_list = recipients

    gui.show()
    sys.exit(app.exec_())



"""Watch an EPICS PV. Send email when it changes from 0 to 1."""

# Copyright (c) 2009-2024, UChicago Argonne, LLC.  See LICENSE file.

import argparse
import datetime
import logging
import os
import socket
import sys
import threading
import time

import epics

from . import PROJECT
from . import __version__ as VERSION
from . import ini_config
from . import mailer

LOG_FILE = f"pvMail-{os.getpid()}.log"
RETRY_INTERVAL_S = 0.2
CHECKPOINT_INTERVAL_S = 5 * 60.0
CONNECTION_TEST_TIMEOUT = 0.5

gui_object = None


class PvMail(threading.Thread):
    """
    Watches an EPICS PV and sends email when value changes from 0 to 1.
    """

    def __init__(self, config=None):
        self.trigger = False
        self.message = "default message"
        self.subject = "pvMail.py"
        self.triggerPV = ""
        self.messagePV = ""
        self.recipients = []
        self.old_value = None
        self.ca_timestamp = None
        self.config = config
        self.running = False
        self.pv = dict(trigger=None, message=None)
        self.pv_cb_index = dict(trigger=None, message=None)

    def basicChecks(self):
        """
        check for valid inputs,
        raise exceptions as discovered,
        otherwise no return result
        """
        if len(self.recipients) == 0:
            msg = "need at least one email address for list of recipients"
            raise RuntimeWarning(msg)
        parts = {"message": self.messagePV, "trigger": self.triggerPV}
        for key, pv in parts.items():
            if len(pv) == 0:
                raise RuntimeWarning(f"no EPICS PV name for PV {key!r}")
            if self.pv[key] is None:
                if self.testConnect(pv, timeout=CONNECTION_TEST_TIMEOUT) is False:
                    raise RuntimeWarning(f"could not connect to {key} PV: {pv}")
            else:
                if not self.pv[key].connected:
                    raise RuntimeWarning(f"could not connect to {key} PV: {pv}")

    def testConnect(self, pvname, timeout=5.0):
        """
        create PV,
        wait for connection,
        return connection state (True | False)

        adapted from PyEpics __createPV() method
        """
        logger("test connect with %s" % pvname)
        retry_interval_s = 0.0001
        start_time = time.time()
        thispv = epics.PV(pvname)
        thispv.connect()
        while not thispv.connected:
            time.sleep(retry_interval_s)
            epics.ca.poll()
            if time.time() - start_time > timeout:
                break
        return thispv.connected

    def do_start(self):
        """start watching for triggers"""
        logger("do_start")
        if not self.running:
            self.basicChecks()
            logger("passed basicChecks(), starting monitors")

            handler_list = [
                ["message", self.messagePV, self.receiveMessageMonitor],
                ["trigger", self.triggerPV, self.receiveTriggerMonitor],
            ]
            for key, pvname, cb in handler_list:
                pv = epics.PV(pvname)
                self.pv[key] = pv
                self.pv_cb_index[key] = pv.add_callback(cb)

            self.old_value = self.pv["trigger"].get()
            self.message = self.pv["message"].get()

            logger("PVs connected")
            self.running = True

    def do_stop(self):
        """stop watching for triggers"""
        logger("do_stop")
        if self.running:
            for key, pv in self.pv.items():
                if pv is not None and pv.connected:
                    pv.remove_callback(self.pv_cb_index[key])
                    pv.disconnect()
                    self.pv[key] = None
                    self.pv_cb_index[key] = None
            logger("PVs disconnected")
            self.running = False

    def do_restart(self):
        """restart watching for triggers"""
        self.do_stop()
        self.do_start()

    def receiveMessageMonitor(self, value, **kw):
        """respond to EPICS CA monitors on message PV"""
        logger("%s = %s" % (self.messagePV, value))
        self.message = value

    def receiveTriggerMonitor(self, value, **kw):
        """respond to EPICS CA monitors on trigger PV"""
        logger("%s = %s" % (self.triggerPV, value))
        # print self.old_value, type(self.old_value), value, type(value)
        if self.old_value == 0 and value == 1:
            self.trigger = True  # set email trigger flag
            pv = self.pv["trigger"]
            # print pv.pvname, value,pv.timestamp
            self.ca_timestamp = pv.timestamp

            # send the message in a different thread
            # SendMessage(self, self.config)
            t = threading.Thread(target=SendMessage, args=(self, self.config))
            t.start()
        self.old_value = value


def SendMessage(pvm, agent_db, reporter=None):
    """
    construct and send the message

    :param obj pvm: instance of PvMail object on which to report
    """

    # print "SendMessage", type(pvm), pvm
    logger("SendMessage")
    pvm.trigger = False  # triggered event received

    agent_db = agent_db or ini_config.Config()
    email_agent_dict = dict(
        sendmail=mailer.sendMail_sendmail, SMTP=mailer.sendMail_SMTP
    )
    emailer = email_agent_dict[agent_db.mail_transfer_agent]

    try:
        _send(emailer, pvm, agent_db, logger=logger)
    except Exception as exc:
        logger("problem sending email: %s", exc)


def getUserName(db):
    u1 = os.environ.get("LOGNAME", None)
    u2 = os.environ.get("USERNAME", None)
    u3 = db.get()["user"]
    return u1 or u2 or u3


def _send(emailer, pvm, agent_db, reporter=None, logger=None):
    pvm.basicChecks()

    pvm.subject = "pvMail.py: " + pvm.triggerPV

    msg = ""  # start with a new message
    msg += "\n\n"
    msg += epics.caget(pvm.messagePV)
    msg += "\n\n"
    username = getUserName(agent_db)
    msg += "user: %s\n" % username
    msg += "host: %s\n" % socket.gethostname()
    msg += "date: %s (UNIX, not PV)\n" % datetime.datetime.now()
    if hasattr(pvm, "ca_timestamp"):
        msg += "CA_timestamp: %d\n" % pvm.ca_timestamp
    else:
        msg += "CA_timestamp: not available\n"
    msg += "program: %s\n" % sys.argv[0]
    msg += "PID: %d\n" % os.getpid()
    msg += "trigger PV: %s\n" % pvm.triggerPV
    msg += "message PV: %s\n" % pvm.messagePV
    msg += "recipients: %s\n" % ", ".join(pvm.recipients)
    pvm.message = msg

    if logger is not None:
        logger("#" * 60)
        logger(pvm.message)
        logger("#" * 60)

    emailer(pvm.subject, msg, pvm.recipients, agent_db.get(), logger=logger)
    logger("message(s) sent")


def logger(message):
    """
    log a message or report from PvMail

    :param str message: words to be logged
    """
    now = datetime.datetime.now()
    name = os.path.basename(sys.argv[0])
    pid = os.getpid()
    text = "(%d,%s,%s) %s" % (pid, name, now, message)
    logging.info(text)


def cli(results, config=None):
    """
    command-line interface to the PvMail class

    :param obj results: default parameters from argparse, see main()
    :param obj config: email configuration from ini_config.Config()
    """
    logging_interval = min(60 * 60, max(5.0, results.logging_interval))
    sleep_duration = min(5.0, max(0.0001, results.sleep_duration))

    pvm = PvMail(config)
    pvm.triggerPV = results.trigger_PV
    pvm.messagePV = results.message_PV
    pvm.recipients = results.email_addresses.strip().split(",")
    checkpoint_time = time.time()
    pvm.do_start()
    while True:  # endless loop, kill with ^C or equal
        if time.time() > checkpoint_time:
            checkpoint_time += logging_interval
            logger("checkpoint")
        if pvm.trigger:
            SendMessage(pvm, config)
            logger("trigger received, sending email")
        time.sleep(sleep_duration)
    # pvm.do_stop()        # this will never be called


def gui(results, config=None):
    """
    graphical user interface to the PvMail class

    :param obj results: default parameters from argparse, see main()
    :param obj config: email configuration from ini_config.Config()
    """

    from . import uic_gui

    uic_gui.main(
        results.trigger_PV,
        results.message_PV,
        results.email_addresses.strip().split(","),
        logger=logger,
        logfile=results.log_file,
        config=config,
    )


def main():
    """parse command-line arguments and choose which interface to use"""
    doc = f"{PROJECT}, v{VERSION}, {__doc__.strip()}"
    parser = argparse.ArgumentParser(description=doc)

    # positional arguments
    # not required if GUI option is selected
    parser.add_argument(
        "trigger_PV",
        action="store",
        nargs="?",
        help="EPICS trigger PV name",
        default="",
    )

    parser.add_argument(
        "message_PV",
        action="store",
        nargs="?",
        help="EPICS message PV name",
        default="",
    )

    parser.add_argument(
        "email_addresses",
        action="store",
        nargs="?",
        help="email address(es), comma-separated if more than one",
        default="",
    )

    # optional arguments
    parser.add_argument(
        "-l",
        action="store",
        dest="log_file",
        help="for logging program progress and comments",
        default=LOG_FILE,
    )

    parser.add_argument(
        "-i",
        action="store",
        dest="logging_interval",
        type=float,
        help="checkpoint reporting interval (s) in log file",
        default=CHECKPOINT_INTERVAL_S,
    )

    parser.add_argument(
        "-r",
        action="store",
        dest="sleep_duration",
        type=float,
        help="sleep duration (s) in main event loop",
        default=RETRY_INTERVAL_S,
    )

    parser.add_argument(
        "-g",
        "--gui",
        action="store_true",
        default=False,
        dest="interface",
        help="Use the graphical rather than command-line interface",
    )

    parser.add_argument("-v", "--version", action="version", version=VERSION)

    results = parser.parse_args()

    addresses = results.email_addresses.strip().split(",")
    interface = {False: "command-line", True: "GUI"}[results.interface]

    agent_db = ini_config.Config()

    results.log_file = results.log_file.strip()

    logging.basicConfig(filename=results.log_file, level=logging.INFO)
    logger("#" * 60)
    logger("startup")
    logger("trigger PV       = " + results.trigger_PV)
    logger("message PV       = " + results.message_PV)
    logger("email list       = " + str(addresses))
    logger("log file         = {results.log_file}")
    logger("logging interval = " + str(results.logging_interval))
    logger("sleep duration   = " + str(results.sleep_duration))
    logger("interface        = " + interface)
    user = os.environ.get("LOGNAME", None) or os.environ.get("USERNAME", None)
    logger("user             = " + user)
    logger("host             = " + socket.gethostname())
    logger("program          = " + sys.argv[0])
    logger("PID              = " + str(os.getpid()))
    logger("PyEpics version  = " + str(epics.__version__))
    logger("config file      = " + agent_db.ini_file)

    if results.interface is False:
        # When the GUI is not selected,
        # ensure the positional arguments are given
        tests = [
            len(results.trigger_PV),
            len(results.message_PV),
            len(" ".join(addresses)),
        ]
        if 0 in tests:
            parser.print_usage()
            sys.exit()

    # call the interface
    {False: cli, True: gui}[results.interface](results)


if __name__ == "__main__":
    main()
